import tensorflow_hub as hub
from tensorflow_text import SentencepieceTokenizer
import tensorflow as tf

class ModelUSE():
    def __init__(self):
        # Ограничение потребляемой памяти моделью
        gpus = tf.config.experimental.list_physical_devices('GPU')
        tf.config.experimental.set_virtual_device_configuration(gpus[0],
                                                                [tf.config.experimental.VirtualDeviceConfiguration(
                                                                    memory_limit=1024)])
        self.model = hub.load("https://tfhub.dev/google/universal-sentence-encoder-multilingual-large/3")

    def encode(self, sentences, batch_size=32, **kwargs):
        embeddings = []
        for i in range(0, len(sentences), batch_size):
            batch_sentences = sentences[i:i+batch_size]
            batch_embeddings = self.model(batch_sentences)
            embeddings.extend(batch_embeddings)
        return embeddings


from rank_bm25 import BM25Okapi
from typing import List, Dict, Tuple, Union
from sentence_transformers import SentenceTransformer, util
from .evaluation_metrics import (
    TopK, AverageLoc, FDARO
)
from .models import ModelUSE
import tensorflow as tf
from tensorflow import convert_to_tensor
import torch

class Bm25:
    """Класс метрики ранжирования bm25"""

    def __init__(self):
        pass

    def name(self):
        return "Bm25"

    def ranking(self, query: str, sentences: List[str], labels: List[int]) -> List[Tuple[float, int]]:
        """
        Функция ранжирования bm25. https://ru.wikipedia.org/wiki/Okapi_BM25

        Parameters
        ------------
        query: `str`
            Список токенов запроса
        sentences: `List[str]`
            Список списков токенов текстов
        labels: `List[int]`
            Список меток текстов

        Returns
        ------------
        `List[Tuple[float, int]]`
            Список оценок релевантности текстов
        """
        tokenized_query = self._encode(query)[0]
        tokenized_sentences = self._encode(sentences)
        bm25 = BM25Okapi(tokenized_sentences)
        scores = bm25.get_scores(tokenized_query)

        scores = self._sorted(scores, labels)
        return scores

    def _sorted(self, scores: List[float], labels: List[int]) -> List[Tuple[float, int]]:
        """
        Функция сортировки оценки и лейблов

        Parameters
        ------------
        scores: `List[float]`
            Массив оценок ранга присвоенных ранкером
        labels: `List[int]`
            Массив меток

        Returns
        ------------
        `List[List]`
            Отсортированный список ранжируемых элементов по релевантности
        """
        return sorted([item for item in zip(scores, labels)], key=lambda x: x[0], reverse=True)

    def _encode(self, sentences: Union[str, List[str]]) -> List[List[str]]:
        """
        Функция для декодирования предложений в последовательность токенов

        Parameters
        ------------
        sentences: `str, List[str]`
            Строка или массив строк для разбиения на токены

        Returns
        ------------
        `List[List[str]]`
            Последовательность токенов
        """
        # tokenized_sentences = self.tokenizer.encode(sentences, return_tensors="pt")
        # TODO: Сделать более умную токенизацию
        if isinstance(sentences, str):
            sentences = [sentences]

        tokenized_sentences = []
        for cur_sent in sentences:
            tokenized_sentences.append(cur_sent.split(" "))
        return tokenized_sentences


class LaBSE:
    """Класс метрики ранжирования LaBSE"""

    def __init__(self) -> None:
        self.model = SentenceTransformer('sentence-transformers/LaBSE')

    def name(self) -> str:
        return "LaBSE"

    def ranking(self, query: str, sentences: List[str], labels: List[int]) -> List[Tuple[float, int]]:
        """
        Функция ранжирования LaBSE

        Parameters
        ------------
        query: `str`
            Строка запроса
        sentences: `List[str]`
            Список строк текстов
        labels: `List[int]`
            Список меток текстов

        Returns
        ------------
        `List[Tuple[float, int]]`
            Список оценок релевантности текстов
        """
        query = self.model.encode(query)
        embeddings = self.model.encode(sentences)
        scores = util.pytorch_cos_sim(query, embeddings).numpy()[0]
        scores = self._sorted(scores, labels)
        return scores

    def _sorted(self, scores: List[float], labels: List[int]) -> List[Tuple[float, int]]:
        """
        Функция сортировки оценки и лейблов

        Parameters
        ------------
        scores: `List[float]`
            Массив оценок ранка присвоенных ранкером
        labels: `List[int]`
            Массив меток

        Returns
        ------------
        `List[Tuple[float, int]]`
            Отсортированный список ранжируемых элементов по релевантности
        """
        return sorted([item for item in zip(scores, labels)], key=lambda x: x[0])

class USE:
    """Класс метрики ранжирования USE"""
    def __init__(self):
        self.model = ModelUSE()

    def name(self) -> str:
        return "USE"

    def ranking(self, query: str, sentences: List[str], labels: List[int]) -> List[Tuple[float, int]]:
        """
        Функция ранжирования LaBSE

        Parameters
        ------------
        query: `str`
            Строка запроса
        sentences: `List[str]`
            Список строк текстов
        labels: `List[int]`
            Список меток текстов

        Returns
        ------------
        `List[Tuple[float, int]]`
            Список оценок релевантности текстов
        """
        # query = self.model.encode(query)
        sentences.append(query)
        embeddings = [x.numpy() for x in self.model.encode(sentences)]
        embeddings = torch.Tensor(embeddings)
        scores = util.pytorch_cos_sim(embeddings[-1], embeddings[:-1]).numpy()[0]
        scores = self._sorted(scores, labels)
        return scores

    def _sorted(self, scores: List[float], labels: List[int]) -> List[Tuple[float, int]]:
        """
        Функция сортировки оценки и лейблов

        Parameters
        ------------
        scores: `List[float]`
            Массив оценок ранка присвоенных ранкером
        labels: `List[int]`
            Массив меток

        Returns
        ------------
        `List[Tuple[float, int]]`
            Отсортированный список ранжируемых элементов по релевантности
        """
        return sorted([item for item in zip(scores, labels)], key=lambda x: x[0])

class RankingMetrics:
    """Класс аккумулирующий все метрики"""
    FAKE_DOC_LABEL: int = -1

    def __init__(self, metrics) -> None:
        # Среднее место фейковых документов в финальной выдаче
        self.average_place_fake_doc = AverageLoc(metrics)
        # Количество случаев когда фейковый документ выше релевантного
        self.fake_doc_above_relevant_one = FDARO(metrics)
        # Количество случаев когда фейковый документ вошел в топ 1
        self.fake_top_k = TopK(metrics)
        # Массив метрик для подсчета
        self.metrics = metrics

    def update(self, query: str, sentences: List[str], labels: List[int]) -> None:
        """
       Функция обновления всех метрик по переданным данным

       Parameters
       ------------
       query: `str`
           Строка запроса
       sentences: `List[str]`
           Список строк текстов
       labels: `List[int]`
           Список меток текстов
       """
        if not isinstance(query, str):
            raise TypeError("The request must be a string!")

        if len(sentences) != len(labels):
            raise ValueError("len(labels) must be equal to len(sentences)")

        for item in sentences:
            if not isinstance(item, str):
                raise TypeError("The sentences must be of the `List[str]` type!")

        for item_label in labels:
            if not isinstance(item_label, int):
                raise TypeError("The labels must be of the `List[int]` type!")

        for cur_metric in self.metrics:
            ranking_list = cur_metric.ranking(query, sentences, labels)
            self.fake_top_k.update(cur_metric.name(), ranking_list, RankingMetrics.FAKE_DOC_LABEL)
            self.fake_doc_above_relevant_one.update(cur_metric.name(), ranking_list, RankingMetrics.FAKE_DOC_LABEL)
            self.average_place_fake_doc.update(cur_metric.name(), ranking_list, RankingMetrics.FAKE_DOC_LABEL)

    def get(self) -> Dict:
        """
        Функция для получения значения всех метрик

        Returns
        ----------
        `Dict`
            Словарь значений метрик
        """
        result = {}
        for key_, value in self.average_place_fake_doc.get().items():
            result[key_] = value

        for key_, value in self.fake_top_k.get().items():
            result[key_] = value

        for key_, value in self.fake_doc_above_relevant_one.get().items():
            result[key_] = value

        return result



from typing import List, Union, Tuple, Dict
import numpy as np

class TopK:
    TOP1 = 1
    TOP3 = 3
    TOP5 = 5

    def __init__(self, ranking_metrics: List) -> None:
        self.metrics = {}
        self.calls_cnt = {}
        self._separator = "_"
        self._top_numbers = [TopK.TOP1, TopK.TOP3, TopK.TOP5]
        for cur_metrics in ranking_metrics:
            for cur_top in self._top_numbers:
                self.metrics[cur_metrics.name() + self._separator + self.name() + str(cur_top)] = 0
                self.calls_cnt[cur_metrics.name() + self._separator + self.name() + str(cur_top)] = 0

    def update(self, metric_name: str,
               ranking_list: List[Union[Tuple[float, int], List[Union[float, int]]]],
               fake_doc_label: Union[int, List[int]] = -1) -> None:
        """
        Функция для обновления значений метрики

        Parameters
        -------------
        metric_name: `str`
            Название метрики
        ranking_list: `List[Union[Tuple[float, int], List[float, int]]]`
            Результат ранжирующей модели
        fake_doc_label: `Union[int, List[int]]`
            Метка или массив меток, принадлежащих фейковым документам
        """
        if isinstance(fake_doc_label, int):
            fake_doc_label = [fake_doc_label]

        _check_update_args(fake_doc_label, ranking_list)

        for top_num in self._top_numbers:
            for item in ranking_list[:min(len(ranking_list), top_num)]:
                if item[1] in fake_doc_label:
                    self.metrics[metric_name + self._separator + self.name() + str(top_num)] += 1

            self.calls_cnt[metric_name + self._separator + self.name() + str(top_num)] += 1

    def get(self) -> Dict[str, float]:
        """
        Функция для получения словаря значений подсчитанных метрик

        Returns
        -------------
        `Dict[str, float]`
            Словарь подсчитанных значений метрик
        """
        result = {}
        for metric_name, value in self.metrics.items():
            metric_name_ = metric_name.split("_")[0]
            top_num = int(metric_name.split('@')[1])
            result[metric_name_ + self._separator + self.name() + str(top_num)] = \
                value / max(1, self.calls_cnt[metric_name_ + self._separator + self.name() + str(top_num)])

        return result

    def name(self) -> str:
        '''
        Функция для получения имени метрики

        Returns
        -------------
        `str`
            Имя метрики
        '''
        return "Top@"


class FDARO:
    """Метрика оценивает как часто фейковый документ находится выше релевантного"""

    def __init__(self, ranking_metrics: List) -> None:
        self._separator = "_"
        self.metrics, self.calls_cnt = {}, {}
        for cur_metric in ranking_metrics:
            self.metrics[cur_metric.name() + self._separator + self.name()] = 0
            self.calls_cnt[cur_metric.name() + self._separator + self.name()] = 0

    def update(self, metric_name: str,
               ranking_list: List[Union[Tuple[float, int], List[Union[float, int]]]],
               fake_doc_label: Union[int, List] = -1) -> None:
        """
        Функция для обновления значений метрики

        Parameters
        -------------
        metric_name: `str`
            Название метрики
        ranking_list: `List[Union[Tuple[float, int], List[float, int]]]`
            Результат ранжирующей модели
        fake_doc_label: `Union[int, List[int]]`
            Метка или массив меток, принадлежащих фейковым документам
        """
        if isinstance(fake_doc_label, int):
            fake_doc_label = [fake_doc_label]

        _check_update_args(fake_doc_label, ranking_list)

        is_first = False
        for item in ranking_list:
            if item[1] > 0:
                break
            elif item[1] in fake_doc_label:
                is_first = True
                break

        if is_first:
            self.metrics[metric_name + self._separator + self.name()] += 1

        self.calls_cnt[metric_name + self._separator + self.name()] += 1

    def get(self) -> Dict[str, float]:
        """
        Функция для получения словаря значений подсчитанных метрик

        Returns
        -------------
        `Dict[str, float]`
            Словарь подсчитанных значений метрик
        """
        result = {}
        for metric_name, value in self.metrics.items():
            metric_name = metric_name.split("_")[0]
            result[metric_name + self._separator + self.name()] = \
                value / max(1, self.calls_cnt[metric_name + self._separator + self.name()])

        return result

    def name(self) -> str:
        '''
        Функция для получения имени метрики

        Returns
        -------------
        `str`
            Имя метрики
        '''
        return "FDARO"


class AverageLoc:
    """Метрика для оценки среднего места фейкового документа"""

    def __init__(self, ranking_metrics) -> None:
        self._separator = "_"
        self.metrics, self.calls_cnt = {}, {}
        for cur_metric in ranking_metrics:
            self.metrics[cur_metric.name() + self._separator + self.name()] = 0
            self.calls_cnt[cur_metric.name() + self._separator + self.name()] = 0

    def update(self, metric_name: str,
               ranking_list: List[Union[Tuple[float, int], List[Union[float, int]]]],
               fake_doc_label: Union[int, list] = -1) -> None:
        """
        Функция для обновления значений метрики

        Parameters
        -------------
        metric_name: `str`
            Название метрики
        ranking_list: `List[Union[Tuple[float, int], List[float, int]]]`
            Результат ранжирующей модели
        fake_doc_label: `Union[int, List[int]]`
            Метка или массив меток, принадлежащих фейковым документам
        """
        if isinstance(fake_doc_label, int):
            fake_doc_label = [fake_doc_label]

        _check_update_args(fake_doc_label, ranking_list)

        is_fake = False
        for ind, item in enumerate(ranking_list):
            if item[1] in fake_doc_label:
                self.metrics[metric_name + self._separator + self.name()] += (ind + 1)
                is_fake = True
        # Если нет фейковых документов в рейтинге, предполагаем что он идет после всех т.е последним
        if not is_fake:
            self.metrics[metric_name + self._separator + self.name()] += len(ranking_list) + 1

        self.calls_cnt[metric_name + self._separator + self.name()] += 1

    def get(self) -> Dict[str, float]:
        """
        Функция для получения словаря значений подсчитанных метрик

        Returns
        -------------
        `Dict[str, float]`
            Словарь подсчитанных значений метрик
        """
        result = {}
        for metric_name, value in self.metrics.items():
            metric_name = metric_name.split("_")[0]
            result[metric_name + self._separator + self.name()] = \
                value / max(1, self.calls_cnt[metric_name + self._separator + self.name()])

        return result

    def name(self) -> str:
        '''
        Функция для получения имени метрики

        Returns
        -------------
        `str`
            Имя метрики
        '''
        return "AverageLoc"


def _check_update_args(fake_doc_label: Union[int, List[int]],
                       ranking_list: List[Union[Tuple[float, int], List[Union[float, int]]]]) -> None:
    '''
    Функция для валидации списка меток и списка для ранжирования

    Parameters
    -------------
    fake_doc_label: `Union[int, List[int]]`
        Метка или списко меток, обозначающих фейковый документ
    ranking_list: `List[Union[Tuple[float, int], List[float, int]]]`
        Список оценок ранжировщика
    '''
    if not isinstance(fake_doc_label, list):
        raise TypeError("The tags of fake documents must be 'int' or 'list!'")
    else:
        for label in fake_doc_label:
            if not isinstance(label, int):
                raise TypeError("The tags of fake documents must be 'int'!")

    if not isinstance(ranking_list, list):
        raise TypeError("The list of ranked elements should be of type 'list'!")
    else:
        for item in ranking_list:
            if not isinstance(item, (tuple, list)) or \
                    not isinstance(item[0], (np.floating, float)) or not isinstance(item[1], int) or \
                    len(item) != 2:
                raise TypeError("The list of ranked elements should contain pairs of elements: rating, label")
